/* 
 * Adaptive spline approximation using picewise constants
 * Author: eckhardt, raab
 *
 * Edited on 30. September 2011, 11:01
 *
 * The software package "Adaptive spline approximation" is 
 * a Beta Version 0.9. 
 * If you have any suggestions or remarks please feel free 
 * to contact our workgroup. This software package is 
 * provides solely for information purposes.  
 * Install and run it on your own responsibility, only!
 *
 */

#define FUNCTION1 sin(2*pi*(x*x*x))    // for computation 
#define STRFUNCTION1 "sin(2 pi (x^3))" // for visual output

#define FUNCTION2 sin(2*pi*x) 
#define STRFUNCTION2 "sin(2 pi x)"  

#define FUNCTION3 sin(2*pi*x*x*x*x)+2*pi*(x-0.1)*(x-0.25)*(x-0.9) 
#define STRFUNCTION3 "sin(2 pi x^4) + 2 pi (x-0.1)*(x-0.25)*(x-0.9)"

#define FUNCTION4 2*pi*(x-0.1)*(x-0.25)*(x-0.9)
#define STRFUNCTION4 "2 pi (x-0.1)*(x-0.25)*(x-0.9)"

#define SAMPLERESOLUTION 1000  // i.e. 1000

#include <stdlib.h>
#include <stdio.h>
#include <fstream>
#include <iostream>
#include <math.h>

using namespace std;

/* global data container */
const int N = SAMPLERESOLUTION;        // number of sample values
const double pi = 3.14159265; 
double functionValues[N+1] = {0.0};    // sampled function values
double deviationValues[N+1] = {0.0};   // sampled deviation values 
double approximateValues1[N] = {0.0};  // approximation 
double epsilon = 0.1;                  // tolerance of approximation
double errEst;                         // for error estimation
char* strfunction;                     // string format of function 


/*
 * samples a given function and 
 * computes sampled deviation values
 */
void setupSampledValues(int functionNumber)
{
    // <editor-fold defaultstate="collapsed" desc="CODE setupSampledValues()">
    
    double x, value;
   
    /* sample function */
    switch(functionNumber)
    {
        case 1:
        {
            strfunction = STRFUNCTION1;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION1;
                functionValues[i] = value;
            }
            break;
        }
        case 2:
        {
            strfunction = STRFUNCTION2;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION2;
                functionValues[i] = value;
            }
            break;
        }
        case 3:
        {
            strfunction = STRFUNCTION3;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION3;
                functionValues[i] = value;
            }
            break;
        }
        default: //case 4:
        {
            strfunction = STRFUNCTION4;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION4;
                functionValues[i] = value;
            }
            break;
        }
        //default: cout << "Error. Unknown formula." << endl;
        //break;
    } 
    
    /* compute sampled deviation */
    for (int k = 0; k < N; k++)
    {
        //deviationValues[k-1] = (N/(2*pi)) * (functionValues[k] - functionValues[k-1]);
        deviationValues[k] = (N-1) * (functionValues[k+1] - functionValues[k]);
    }
    deviationValues[N] = deviationValues[N - 1];
  
    // </editor-fold>
}


/*
 * Computes minimum of the sampled values 
 * in the range determined by j and k. 
 */
double computeMin(int j, int k) 
{
    // <editor-fold defaultstate="collapsed" desc="CODE computeMin()">
    
    /* transform j,k into coordinates of the sampled values */
    int size = (int)(pow(2, (-1)*j) * N);
    int from = k*size;
    int to = (k+1)*size;
 
    /* compute minimum */
    double min = functionValues[from];
    //double dnorm = 0;
    for (int i = from; i < to; i++)
    {
        if (min > functionValues[i])
        {
            min = functionValues[i];
        }
    }
    
    return (min);
    
    // </editor-fold>
}


/*
 * Computes maximum of the sampled values 
 * in the range determined by j and k. 
 */
double computeMax(int j, int k) 
{
    // <editor-fold defaultstate="collapsed" desc="CODE computeMax()">
    
    /* transform j,k into coordinates of the sampled values */
    int size = (int)(pow(2, (-1)*j) * N);
    int from = k*size;
    int to = (k+1)*size;
 
    /* compute maximum */
    double max = functionValues[from];
    for (int i = from; i < to; i++)
    {
        if (max < functionValues[i])
        {
            max = functionValues[i];
        }
    }
    
    return (max);
    
    // </editor-fold>
}


/*
 * The necessary computations for the local error estimation
 * are computed with use of the first deviation.
 * Update of approximateValues[] with current result. 
 */
double errorEstimate1(int j, int k)
{
    // <editor-fold defaultstate="collapsed" desc="CODE errorEstimate1()">
    
    /* transform j,k into coordinates of the sampled values */
    int size = (int)(pow(2, (-1)*j) * N);
    int from = k*size;
    int to = (k+1)*size;
 
    /* compute minimum and maximum */
    double min = computeMin(j,k);
    double max = computeMax(j,k);
    
    /* compute norm, error and update results */
    double dnorm = 0;
    for (int i = from; i < to; i++)
    {
        dnorm += fabs(deviationValues[i]);
        approximateValues1[i] = (max + min) / 2.0;    
    }

    //double errEst = ((to - from)/double(N) * dnorm);
    double errEst = (dnorm/double(N));
    return (errEst);

    // </editor-fold>
}


/*
 * write approximation data to file 
 */
void writeOutput(ofstream &output_filestream)
{
    // <editor-fold defaultstate="collapsed" desc="CODE writeOutput()">
    
    /* output x-axis and sampled function values */
    output_filestream << "x = " << 1/double(N) << ":" << 1/double(N) << ":1;\n" 
                      << "f = [";
    
    for (int i = 0; i < N; i++)
    {
        output_filestream << functionValues[i] << " ";
    }
    output_filestream << "];\nd = [" << endl;

    /* output deviation values */
    for (int i = 0; i < N; i++)
    {
        output_filestream << deviationValues[i] << " ";
    }
    output_filestream << "];\na = [" << endl;

    /* output approximation values */
    for (int i = 0; i < N; i++)
    {
        output_filestream << approximateValues1[i] << " ";
    }
    // output_filestream << "];\n b = [";
    // for (int i = 0; i < N; i++)
    // {
    //     output_filestream << approximateValues2[i] << " ";
    // }

    output_filestream << "];\npause(0.01);\n"
                      << "if (~ishandle(figHandle)), return, end\n"
                      << "figure(figHandle);\nclf(figHandle);\n"
                      << "plot(x,f,'k', x,a,'r');\n"
                      << "legend('function', 'approximation');\n"
                      << "title('f(x) = " << strfunction << ", tolerance = "
                      << epsilon << "');\n";
    
    // </editor-fold>
}


/*
 * recursive call of errorEstimate() to compute overall approximation
 */
void computeApproximation(int j, int k, ofstream &output_filestream)
{
    // <editor-fold defaultstate="collapsed" desc="CODE computeApproximation()">
    
    errEst = errorEstimate1(j, k);

    writeOutput(output_filestream);
    
    //if ((errEst1 > epsilon)) //&& (pow(2, (-1)*j) > 1/double(N)))
    if ((errEst > epsilon)) //&& (pow(2, (-1)*j) > 1/double(N)))  
    {
        /* divide interval */
        //cout << "% divide (" << j << ", " << k << ")" << endl;
        computeApproximation(j+1, 2*k, output_filestream);
        computeApproximation(j+1, 2*k + 1, output_filestream);
    }

    // </editor-fold>
}


/*
 * first parameter determines function to be approximated
 * second parameter the tolerance epsilon
 */
int main(int argc, char** argv)
{
    int functionNumber = atoi(argv[1]);  // which function to approximate
    epsilon = atof(argv[2]);             // approximation tolerance
    
    setupSampledValues(functionNumber);
    
    cout << "% Approximation by picewise constants of " << endl
         << "% f(x) = " << strfunction << endl
         << "% tolerance = " << epsilon << endl
         << "% sample values = " << N << endl;

    /* create output file */
    char output_filename[250];
    sprintf(output_filename, "output_figure_picewise_constants_function%d.m", functionNumber);
    ofstream output_filestream(output_filename);
    output_filestream << "% Approximation by picewise constants of "
                      << "% f(x) = " << strfunction << ".\n"
                      << "% tolerance = " << epsilon
                      << "% sample values = " << N
                      << "\nfigHandle = figure; hold on;" << endl;
                      //<< "legend('function', 'approximation');\n"
                      //<< "title('f(x) = " << strfunction << ", tolerance = "
                      //<< epsilon << "');\n";
    
    /* start to compute the approximation */
    computeApproximation(0, 0, output_filestream);   // on [0, 1]
    
    output_filestream << "\nxlabel('Done.');" << endl;
    output_filestream.close();

    cout << "Output written to file: " << output_filename << endl;
    
    return (EXIT_SUCCESS);
}
