/* 
 * Adaptive spline approximation using picewise linear functions
 * Author: eckhardt, raab
 *
 * Created on 02. August 2011, 10:20
 *
 * The software package "Adaptive spline approximation" is 
 * a Beta Version 0.9. 
 * If you have any suggestions or remarks please feel free 
 * to contact our workgroup. This software package is 
 * provides solely for information purposes.  
 * Install and run it on your own responsibility, only!
 *
 */

#define FUNCTION1 sin(2*pi*(x*x*x))    // for computation 
#define STRFUNCTION1 "sin(2 pi (x^3))" // for visual output

#define FUNCTION2 sin(2*pi*x) 
#define STRFUNCTION2 "sin(2 pi x)"  

#define FUNCTION3 sin(2*pi*x*x*x*x)+2*pi*(x-0.1)*(x-0.25)*(x-0.9) 
#define STRFUNCTION3 "sin(2 pi x^4) + 2 pi (x-0.1)*(x-0.25)*(x-0.9)"

#define FUNCTION4 2*pi*(x-0.1)*(x-0.25)*(x-0.9)
#define STRFUNCTION4 "2 pi (x-0.1)*(x-0.25)*(x-0.9)"

#define SAMPLERESOLUTION 1000  // i.e. 1000

#include <stdlib.h>
#include <stdio.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include <string.h>

using namespace std;

/* global data container */
const int N = SAMPLERESOLUTION;             // number of sample values
const double pi = 3.14159265;             
double functionValues[N+1] = {0.0};         // sampled function values
double deviationValues[N+1] = {0.0};        // sampled deviation values
double secondDeviationValues[N+1] = {0.0};  // sampled second deviation values
int essentialPoint[N+1] ;                   // if entry i == 1 then use point i 
double epsilon = 0.1;                       // tolerance of approximation
double errEst;                              // for error estimation
char* strfunction;                          // string format of function 


/*
 * samples a given function and 
 * computes sampled deviation values 
 */
void setupSampledValues(int functionNumber)
{
    // <editor-fold defaultstate="collapsed" desc="CODE setupSampledValues()">
    
    /* 
     * at the beginning the function is approximated 
     * by the linear function passing through the first 
     * and the last sampled point
     */
    essentialPoint[0]=1;
    essentialPoint[N]=1;

    double x, value;
    
    /* sample function */
    switch(functionNumber)
    {
        case 1: 
        {
            strfunction = STRFUNCTION1;   
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION1;
                functionValues[i] = value;
            }
            break;        
        }
        case 2: 
        {
            strfunction = STRFUNCTION2;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION2;
                functionValues[i] = value;
            }
            break;        
        }
        case 3: 
        {
            strfunction = STRFUNCTION3;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION3;
                functionValues[i] = value;
            }
            break;        
        }
        default: //case 4: 
        {
            strfunction = STRFUNCTION4;
            for (int i = 0; i < N; i++)
            {
                x = i/double(N);
                value = FUNCTION4;
                functionValues[i] = value;
            }
            break;        
        }
        //default: cout << "Unknown formular";
        //break;        
    }   
   
    /* compute sampled deviation */
    for (int k = 0; k < N; k++)
    {
        //deviationValues[k-1] = (N/(2*pi)) * (functionValues[k] - functionValues[k-1]);
        deviationValues[k] = (N-1) * (functionValues[k+1] - functionValues[k]);
    }
    deviationValues[N] = deviationValues[N - 1];

    /*
     * compute second deviation values  
     * (for the first value there is no left point 
     * to compute the second deviation)
     */
    secondDeviationValues[0] = pow((N-1),2) * (functionValues[1] - functionValues[0]);   
    for (int k = 1; k < N; k++)
    {
        //deviationValues[k-1] = (N/(2*pi)) * (functionValues[k] - functionValues[k-1]);
        secondDeviationValues[k] = pow((N-1),2) * (functionValues[k+1] + functionValues[k-1]-2*functionValues[k]);
    }
    secondDeviationValues[N] = secondDeviationValues[N - 1];
   
    // </editor-fold>
}


/*
 * The necessary computations for the local error estimation, 
 * computed by means of the second deviation.
 * Update of essentialPoint[] with current result. 
 */
double errorEstimate2(int from, int to)
{
    // <editor-fold defaultstate="collapsed" desc="CODE errorEstimate2()">
    
    /* transform j,k into coordinates of the sampled values*/
    //  int size = (int)(pow(2, (-1)*j) * N);
    //  int from = k*size;
    // int to = (k+1)*size;
   
    /* compute second deviation norm */
    double dnorm = 0;
    for (int i = from; i < to; i++)
    {
        dnorm += fabs(secondDeviationValues[i]);           
    }
  
    /* update result */
    essentialPoint[from] = 1;
    essentialPoint[to] = 1;

    //double errEst = ((to - from)/double(N) * dnorm);
    double errEst = (dnorm/double(N));
    return (errEst);

    // </editor-fold>
}


/*
 * write computed approximation to file 
 */
void writeOutput(ofstream &output_filestream)
{
    // <editor-fold defaultstate="collapsed" desc="CODE writeOutput()">
    
    /* output x-axis and sampled function values */
    output_filestream << "x = " << 1/double(N) << ":" << 1/double(N) << ":1;\n"
                      << "f = [";
    
    for (int i = 0; i < N; i++)
    {
        output_filestream << functionValues[i] << " ";
    }
    output_filestream << "];\nz = [" << endl;

    /*output the x-Values of the approximation*/ 
    if(essentialPoint[0] == 1) 
    {
        output_filestream << 0.0 << " ";
    }
    
    for (int i = 1; i <= N; i++)
    {
        if (essentialPoint[i] == 1) 
        {   
            output_filestream << i/double(N) << " ";
        }
    }  
    output_filestream << "];\nd  = [" << endl;

    /* output deviation values */
    
    for (int i = 0; i <= N; i++)
    {
        output_filestream << deviationValues[i] << " ";
    }
    output_filestream << "];\na = [" << endl;

    /* output approximation values */
    for (int i = 0; i <= N; i++)
    {
        if (essentialPoint[i] == 1) 
        {  
            output_filestream << functionValues[i] << " ";
        }
    }
    output_filestream << "];\npause(0.01);\n"
                      << "if (~ishandle(figHandle)), return, end\n"
                      << "figure(figHandle);\nclf(figHandle);\n"          
                      << "plot(x,f,'k',z,a,'r');\n"
                      << "legend('function', 'approximation');\n"
                      << "title('f(x) = " << strfunction << ", tolerance = "
                      << epsilon << "');" << endl;
    
    // </editor-fold>
}


/*
 * recursive call of errorEstimate() to compute overall approximation
 */
void computeApproximation(int from, int to, ofstream &output_filestream)
{
    // <editor-fold defaultstate="collapsed" desc="CODE computeApproximation()">
    
    errEst = errorEstimate2(from, to);
      
    int middle = (from + ( (double(to)-double(from))/2.0 ) );
    // cout << "(from, to) = (" << from << "," << to << ")" << endl;
    // cout << "middle = " << middle << endl;
    // cout << "Middle = " << middle;
    
    writeOutput(output_filestream);

    if ((errEst > epsilon) && (middle-from) > 1 ) //&& (pow(2, (-1)*j) > 1/double(N)))  
    {
        computeApproximation(from,middle, output_filestream);
        computeApproximation(middle, to, output_filestream);
    }
    
    // </editor-fold>
}


/*
 * first parameter determines function to be approximated
 * second parameter the tolerance epsilon
 */
int main(int argc, char** argv)
{
    int functionNumber = atoi(argv[1]);  // which function to approximate
    epsilon = atof(argv[2]);             // approximation tolerance
    
    setupSampledValues(functionNumber);
    
    cout << "% Approximation by picewise linear functions of " << endl
         << "% f(x) = " << strfunction << endl
         << "% tolerance = " << epsilon << endl
         << "% sample values = " << N << endl;

    /* create output file */
    char output_filename[250];
    sprintf(output_filename, "output_figure_picewise_linear_function%d.m", functionNumber);
    ofstream output_filestream(output_filename);
    output_filestream << "% Approximation by picewise linear functions of "
                      << "% f(x) = " << strfunction << ".\n"
                      << "% tolerance = " << epsilon
                      << "% sample values = " << N
                      << "\nfigHandle = figure; hold on;" << endl;
                      //<< "legend('function', 'approximation');\n"
                      //<< "title('f(x) = " << strfunction << ", tolerance = "
                      //<< epsilon << "');" << endl;
    
    
    /* start to compute the approximation */
    computeApproximation(0, N, output_filestream);   // on [0, 1]

    //refresh(findobj('type','figure','name','Adaptive_spline_approximation'));
    output_filestream << "\nxlabel('Done.');" << endl;
    output_filestream.close();

    cout << "Output written to file: " << output_filename << endl;
    
    return (EXIT_SUCCESS);
}
